#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################
import torch
torch.manual_seed(0)
from helper_code import *
import numpy as np, scipy as sp,os
#from sklearn.impute import SimpleImputer
#from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, TensorDataset
from architecture import ResNet18, ResBlock
from scipy import signal
from torch.autograd import Variable
################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):

    frame_len = 4096*4
    bs = 32
    num_epochs = 35

    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)
    ######################################################################################
    # Split Patient Files into Train, Validation and Test Set
    patient_files_train, patient_files_validation = train_test_split(patient_files, test_size=0.3, shuffle = True, random_state =0)

    #______________________________________________________
    # Load all patients with labels. 
    #______________________________________________________
    recordings_p_train, labels_murmur_p_train, labels_outcome_p_train = load_data(data_folder, patient_files_train)
    recordings_p_validation, labels_murmur_p_validation, labels_outcome_p_validation  = load_data(data_folder, patient_files_validation)

    #______________________________________________________
    # Prepare data frames and labels.
    #______________________________________________________
    print("**************************************************")
    print("Frame Level")
    print("**************************************************")
    X_train, y1_train, y2_train = prepare_data(recordings_p_train, labels_murmur_p_train, labels_outcome_p_train, frame_len)
    X_val, y1_val, y2_val = prepare_data(recordings_p_validation,  labels_murmur_p_validation, labels_outcome_p_validation, frame_len)
    
    #______________________________________________________
    # Data Loaders
    #______________________________________________________
    train_ds = TensorDataset(X_train, y1_train, y2_train)
    train_dl = DataLoader(train_ds,batch_size=bs,shuffle=True)

    val_ds = TensorDataset(X_val, y1_val, y2_val)
    val_dl = DataLoader(val_ds,batch_size=bs,shuffle=True)

    train(num_epochs, train_dl, val_dl, model_folder)
    print('Finished Training')
    ########################################################################################
    if verbose >= 1:
        print('Done.')

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):

    model_path = os.path.join(model_folder, 'model_SelfONN.pth')
    model = ResNet18(1, ResBlock)
    model.load_state_dict(torch.load(model_path))

    return model

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):

    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']

    frame_len = 4096*4

    pcg = []

    for pcg_r in recordings:
        pcg_r = bandpass_filter(pcg_r)
        rec_len = len(pcg_r)
        if rec_len < frame_len:
            pcg_r = extend_ts(pcg_r,frame_len)
            rec_len = frame_len
        tot_frames = rec_len//frame_len
        for f in range(tot_frames):
            # Start of frame in whole recording.
            st = f*frame_len
            # End of window
            end = st + frame_len
            # If ecg window is non zero. Normalize it. 
            pcg_f = pcg_r[st:end]
            pcg_f = signal.resample(pcg_f,frame_len//4)
            if pcg_f.any():
                pcg_f = (pcg_f - pcg_f.mean())/pcg_f.std()
                pcg.append(pcg_f) 
            # All zero pcg window.
            else:
                print("EMPTY PCG")

    pcg = np.expand_dims(pcg,axis=1)
    pcg = torch.tensor(pcg).float()

    model.eval()
    murmur_probs, outcome_probs, _ = model(pcg)

    murmur_probs = murmur_probs.cpu().detach()
    outcome_probs = outcome_probs.cpu().detach()

    Acc_0 = sum(np.argmax(murmur_probs, axis=1) == 0)
    if Acc_0/len(murmur_probs) > 0.2:
        murmur_preds = 0
    else:
        murmur_preds = murmur_probs.mean(axis=0).argmax()

    Acc_0 = sum(np.argmax(outcome_probs, axis=1) == 0)
    if Acc_0/len(outcome_probs) > 0.2:
        outcome_preds = 0
    else:
        outcome_preds = outcome_probs.mean(axis=0).argmax()

    murmur_probs = murmur_probs.mean(axis=0)
    outcome_probs = outcome_probs.mean(axis=0)

    # Choose label with higher probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    murmur_labels[murmur_preds] = 1


    # Choose label with higher probability.
    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    outcome_labels[outcome_preds] = 1

    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probs, outcome_probs))


    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Extract features from the data.
def get_features(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV', 'PhC']
    num_recording_locations = len(recording_locations)
    recording_features = np.zeros((num_recording_locations, 4), dtype=float)
    num_locations = len(locations)
    num_recordings = len(recordings)
    if num_locations==num_recordings:
        for i in range(num_locations):
            for j in range(num_recording_locations):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i])>0:
                    recording_features[j, 0] = 1
                    recording_features[j, 1] = np.mean(recordings[i])
                    recording_features[j, 2] = np.var(recordings[i])
                    recording_features[j, 3] = sp.stats.skew(recordings[i])
    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)


# Load patient data and labels based on given patient file paths.
def load_data(data_folder, patient_files):

    num_patient_files = len(patient_files)

    # Murmur Class mapping.
    classes_murmur = ['Present', 'Unknown', 'Absent']
    # Outcome Class mapping
    classes_outcome = ['Abnormal', 'Normal']

    labels_murmur_p = []
    labels_outcome_p = []
    recordings_p = []

    for i in range(num_patient_files):

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)

        # Extract murmur Label.
        label_murmur = get_murmur(current_patient_data) 
        if label_murmur in classes_murmur:
            current_label_murmur = classes_murmur.index(label_murmur)

        # Extract outcome Label.
        label_outcome = get_outcome(current_patient_data) 
        if label_outcome in classes_outcome:
            current_label_outcome = classes_outcome.index(label_outcome)

        fs = get_frequency(current_patient_data)

        if fs != 4000:
            print(fs)

        # Per Patient Recordings and lables. 
        labels_murmur_p.append(current_label_murmur)
        labels_outcome_p.append(current_label_outcome)
        recordings_p.append(current_recordings) 

        #auscultation_locs = get_locations(current_patient_data)
        #pat_number = patient_files[i].split('\\')[-1].split('.')[0]

    print("Total patients : ",num_patient_files)
    print("# Murmur Labels : ", np.unique(labels_murmur_p, return_counts=True)[1])
    print("# Outcome Labels : ", np.unique(labels_outcome_p, return_counts=True)[1])
    print("________________________________________________")

    return recordings_p, labels_murmur_p, labels_outcome_p


def prepare_data(recordings_p, labels_murmur_p, labels_outcome_p, frame_len):

    pcg_all = []
    murmur_all = []
    outcome_all = []

    #PCG Patient --> PCG Recording --> PCG Frame
    for i, pcg_p in enumerate(recordings_p):
        for pcg_r in pcg_p:
            pcg_r = bandpass_filter(pcg_r)
            rec_len = len(pcg_r)
            if rec_len < frame_len:
                pcg_r = extend_ts(pcg_r,frame_len)
                rec_len = frame_len

            tot_frames = rec_len//frame_len
            for f in range(tot_frames):
                # Start of frame in whole recording.
                st = f*frame_len
                # End of window
                end = st + frame_len
                pcg_f = pcg_r[st:end]
                if pcg_f.any():
                    pcg_f = signal.resample(pcg_f,frame_len//4)
                    pcg_f = (pcg_f - pcg_f.mean())/pcg_f.std()
                    pcg_all.append(pcg_f) 
                    murmur_all.append(labels_murmur_p[i])
                    outcome_all.append(labels_outcome_p[i])

                # All zero pcg window.
                else:
                    print("EMPTY PCG")
    
    murmur_all = np.hstack(murmur_all)
    outcome_all = np.hstack(outcome_all)

    pcg_all = np.vstack(pcg_all)

    pcg_all = np.expand_dims(pcg_all,axis=1) # [b,c,w]

    pcg_all = torch.tensor(pcg_all).float()
    murmur_all = torch.tensor(murmur_all).long()
    outcome_all = torch.tensor(outcome_all).long()



    print("Total Frames : ",len(pcg_all))
    print("# Murmur Labels : ", np.unique(murmur_all, return_counts=True)[1])
    print("# Outcome Labels : ", np.unique(outcome_all, return_counts=True)[1])
    print("________________________________________________")


    return pcg_all, murmur_all, outcome_all



#####################################################################
#                       Training Codes
#####################################################################
# Function to test the model with the test dataset and print the accuracy for the test images
def get_validationLoss(model, device, test_loader, loss_fn):
    
    model.eval()
    validation_loss = 0.0
    
    with torch.no_grad():
        for data in test_loader:
            x1, labels_murmur, labels_outcome = data

            x1 = Variable(x1.to(device))
            labels_murmur = Variable(labels_murmur.to(device))
            labels_outcome = Variable(labels_outcome.to(device))

            # predict classes using images from the training set
            outputs_murmur, outputs_outcome, _ = model(x1)
            # compute the loss based on model output and real labels
            loss_murmur = loss_fn(outputs_murmur, labels_murmur)
            loss_outcome = loss_fn(outputs_outcome, labels_outcome)


            loss = (loss_murmur + loss_outcome )/2


            validation_loss += loss.item()     # extract the loss value
            
            # the label with the highest energy will be our prediction
            #_, predicted = torch.max(outputs.data, 1)
            #total += labels.size(0)
            #accuracy += (predicted == labels).sum().item()
    
    validation_loss = validation_loss/len(test_loader)
    # compute the accuracy over all test images
    #accuracy = accuracy / total
    return(validation_loss)


# Training function. We simply have to loop over our data iterator and feed the inputs to the network and optimize.
def train(num_epochs, train_loader, validation_loader, model_folder):

    from torch.optim import Adam
    import torch.nn as nn

    model = ResNet18(1, ResBlock)

    # Define the loss function with Classification Cross-Entropy loss and an optimizer with Adam optimizer
    loss_fn = nn.CrossEntropyLoss()
    optimizer = Adam(model.parameters(), lr=0.001)

    # Model Training
    best_loss = 1e9

    # Define your execution device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("The model will be running on", device, "device")
    # Convert model parameters and buffers to CPU or Cuda
    model.to(device)

    for epoch in range(num_epochs):  # loop over the dataset multiple times
        
        train_loss = 0.0

        for i, (x1, labels_murmur, labels_outcome ) in enumerate(train_loader, 0):
            
            # get the inputs
            x1 = Variable(x1.to(device))
            labels_murmur = Variable(labels_murmur.to(device))
            labels_outcome = Variable(labels_outcome.to(device))

            # zero the parameter gradients
            optimizer.zero_grad()
            # predict classes using images from the training set
            outputs_murmur, outputs_outcome, _ = model(x1)
            # compute the loss based on model output and real labels
            loss_murmur = loss_fn(outputs_murmur, labels_murmur)
            loss_outcome = loss_fn(outputs_outcome, labels_outcome)

            loss = (loss_murmur + loss_outcome)/2

            # backpropagate the loss
            loss.backward()
            # adjust parameters based on the calculated gradients
            optimizer.step()

            # Let's print statistics for every 1,000 images
            train_loss += loss.item()     # extract the loss value


        # Compute and print the average accuracy fo this epoch when tested over all 10000 test images
        validation_loss = get_validationLoss(model, device, validation_loader, loss_fn)
        print("_____________________________________________________")
        print('Epoch : ', epoch+1,)
        print('Training Loss : ', round(train_loss/len(train_loader),3))
        print('Validation Loss : ', round(validation_loss,3))
        print("_____________________________________________________")
        
        # we want to save the model if the accuracy is the best
        if validation_loss < best_loss:
            print("Saving model at epoch : ", epoch+1)
            save_challenge_model(model_folder, model)
            best_loss = validation_loss


# Save your trained model.
def save_challenge_model(model_folder, model):
    
    model_path = os.path.join(model_folder, 'model_SelfONN.pth')
    torch.save(model.state_dict(), model_path)


def bandpass_filter(data, lowcut = 25, highcut = 400, signal_freq = 4000, filter_order = 1):

    from scipy.signal import butter, lfilter
    """
    Method responsible for creating and applying Butterworth filter.
    :param deque data: raw data
    :param float lowcut: filter lowcut frequency value
    :param float highcut: filter highcut frequency value
    :param int signal_freq: signal frequency in samples per second (Hz)
    :param int filter_order: filter order
    :return array: filtered data
    """
    nyquist_freq = 0.5 * signal_freq
    low = lowcut / nyquist_freq
    high = highcut / nyquist_freq
    b, a = butter(filter_order, [low, high], btype="band")
    y = lfilter(b, a, data)
    return y


def extend_ts(ts, length):
    extended = np.zeros(length)
    siglength = np.min([length, ts.shape[0]])
    extended[:siglength] = ts[:siglength]
    return extended 